PROGRAM Contml(infile, OUTPUT);
   (* version 2.5 *)

CONST maxsp = 20;   (* maximum number of species *)
   maxsp2 = 38;     (* 2 x maxsp - 2 *)
   maxchr = 20;     (* maximum number of characters *)
   smoothings = 4;  (* number of passes through smoothing algorithm *)
   namelength = 10; (* number of characters max. in species name *)

TYPE phenotype = ARRAY [1..maxchr] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;

VAR infile : TEXT;
   numsp, numsp1, numsp2, loci, totalleles, df, outgrno : INTEGER;
   alleles : ARRAY[1..maxchr] OF INTEGER;
   x : ARRAY[1..maxsp] OF phenotype;
   names : ARRAY[1..maxsp] OF naym;
   brief, contchars, outgropt, usertree, didreroot : BOOLEAN;

   PROCEDURE getinput;

      PROCEDURE getnums;
      BEGIN (* getnums *)
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT,' Continuous character Maximum Likelihood');
         WRITELN(OUTPUT, ' method version 2.5');
         WRITELN(OUTPUT);
         READ(infile, numsp, loci);
         WRITELN(OUTPUT, ' ', numsp:4, ' Populations, ', loci:4,
           ' Loci');
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         VAR ch : CHAR;
      BEGIN (* getoptions *)
         brief := FALSE;
         contchars := FALSE;
         usertree := FALSE;
         outgropt := FALSE;
         outgrno := 1;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            IF ch <> ' '
            THEN BEGIN
               IF ORD(ch) > ORD('Z')
               THEN ch := CHR(ORD(ch)+ORD('Z')-ORD('z'));
               CASE ch OF
                  'B' : brief := TRUE;
                  'C' : contchars := TRUE;
                  'O' : outgropt := TRUE;
                  'U' : usertree := TRUE
                  END;
               END;
            END;
         IF outgropt
         THEN BEGIN
            READLN(infile);
            READ(infile, ch, outgrno);
            END;
      END; (* getoptions *)

      PROCEDURE getalleles;
         VAR i : INTEGER;
      BEGIN (* getalleles *)
         IF contchars
         THEN BEGIN
            totalleles := loci;
            df := loci;
            END
         ELSE BEGIN
            totalleles := 0;
            READLN(infile);
            WRITELN(OUTPUT, ' ');
            WRITELN(OUTPUT, ' ', 'Numbers of alleles at the loci:');
            WRITELN(OUTPUT, ' ', '------- -- ------- -- --- -----');
            WRITELN(OUTPUT, ' ');
            FOR i := 1 TO loci DO BEGIN
               IF EOLN(infile)
               THEN READLN(infile);
               READ(infile, alleles[i]);
               totalleles := totalleles + alleles[i];
               WRITE(OUTPUT, alleles[i]:4);
               END;
            df := totalleles - loci;
            END;
         WRITELN(OUTPUT, ' ');
      END; (* getalleles *)

      PROCEDURE getdata;
         VAR i, j, k, l, m, n : INTEGER;
            sum : REAL;
      BEGIN (* getdata *)
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' Name                       Phenotypes');
         WRITELN(OUTPUT, ' ----                       ----------');
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO numsp DO BEGIN
            WRITE(OUTPUT, ' ');
            READLN(infile);
            FOR j := 1 TO namelength DO BEGIN
               READ(infile, names[i, j]);
               WRITE(OUTPUT, names[i, j]);
               END;
            m := 1;
            FOR j := 1 TO loci DO BEGIN
               sum := 0.0;
               IF contchars
               THEN n := 1
               ELSE n := alleles[j] - 1;
               FOR k := 1 TO n DO BEGIN
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, x[i, m]);
                  sum := sum + x[i, m];
                  WRITE(OUTPUT, x[i, m]:10:5);
                  IF (j MOD 6) = 0 THEN BEGIN
                     WRITELN(OUTPUT);
                     FOR l := 1 TO namelength DO WRITE(OUTPUT, ' ');
                     WRITE(OUTPUT, '   ');
                     END;
                  IF NOT contchars
                  THEN BEGIN
                     IF x[i, m] >= 1.0E-6
                     THEN x[i, m] := SQRT(x[i, m]);
                     END;
                  m := m + 1;
                  END;
               IF NOT contchars
               THEN BEGIN
                  x[i, m] := 1.0 - sum;
                  IF x[i, m] >= 1.0E-6
                  THEN x[i, m] := SQRT(x[i, m]);
                  m := m + 1;
                  END;
               END;
            WRITELN(OUTPUT);
            END;
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      getoptions;
      getalleles;
      getdata;
   END; (* getinput *)

   PROCEDURE maketree;

      TYPE nodeptr = ^node;
         node = RECORD
            next, back : nodeptr;
            tip        : BOOLEAN;
            number     : INTEGER;
            view       : phenotype;
            nayme      : naym;
            v, deltav  : REAL;
            bigv, dist : REAL;
            xcoord,
             ycoord    : INTEGER
            END;
         tree = RECORD
            nodep      : ARRAY[1..maxsp2] OF nodeptr;
            likelihood : REAL;
            start      : nodeptr
            END;

      VAR curtree, priortree, bestree : tree;
         nextsp, numtrees, which : INTEGER;
         succeeded : BOOLEAN;

      PROCEDURE setuptree (VAR a : tree);
        VAR i, j : INTEGER;
           p, q : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               NEW(nodep[i]);
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               q := NIL;
               FOR j := 1 TO 3 DO BEGIN
                  NEW(p);
                  p^.tip := FALSE;
                  p^.number := i;
                  p^.next := q;
                  q := p;
                  END;
               p^.next^.next^.next := p;
               nodep[i] := p;
               END;
            likelihood := -99999.0;
            start := nodep[1];
            END;
      END; (* setuptree *)

      PROCEDURE hookup (p, q : nodeptr);
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      FUNCTION evaluate( VAR t : tree ) : REAL;
         VAR sum : REAL;

         PROCEDURE sumlikely (p, q : nodeptr);
            VAR i : INTEGER;
               sumsq, vee : REAL;
         BEGIN (* sumlikely *)
            IF NOT p^.tip
            THEN sumlikely (p^.next^.back, p^.next^.next^.back);
            IF NOT q^.tip
            THEN sumlikely (q^.next^.back, q^.next^.next^.back);
            sumsq := 0.0;
            FOR i := 1 TO totalleles DO
               sumsq := sumsq + SQR(p^.view[i] - q^.view[i]);
            vee := p^.v + q^.v + p^.deltav + q^.deltav;
            sum := sum - df*LN(vee)/2.0 - sumsq/(2.0*vee);
         END; (* sumlikely *)

      BEGIN (* evaluate *)
         sum := 0.0;
         sumlikely (t.start, t.start^.back);
         t.likelihood := sum;
         evaluate := sum;
      END; (* evaluate *)

      PROCEDURE update (p : nodeptr);
         VAR q : nodeptr;
            negatives : BOOLEAN;

         PROCEDURE makedists (p : nodeptr);
            VAR i : INTEGER;

            FUNCTION distance (p, q : nodeptr) : REAL;
               VAR i : INTEGER;
                  sum : REAL;
            BEGIN (* distance *)
               sum := 0.0;
               FOR i := 1 TO totalleles DO
                  sum := sum + SQR(p^.view[i] - q^.view[i]);
               distance := sum;
            END; (* distance *)

         BEGIN (* makedists *)
            FOR i := 1 TO 3 DO BEGIN
               q := p^.next;
               p^.dist := distance (p^.back, q^.back);
               p := q;
               END;
         END; (* makedists *)

         PROCEDURE makebigv (p : nodeptr);
            VAR i : INTEGER;
               temp, q, r : nodeptr;
         BEGIN (* makebigv *)
            q := p^.next;
            r := q^.next;
            negatives := FALSE;
            FOR i := 1 TO 3 DO BEGIN
               p^.bigv := (p^.dist + r^.dist - q^.dist)/(2*df);
               IF p^.bigv < p^.back^.deltav THEN negatives := TRUE;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* makebigv *)

         PROCEDURE correctv (p : nodeptr);
            VAR q, r, temp : nodeptr;
               i, j : INTEGER;
               f1, f2, vtot : REAL;
         BEGIN (* correctv *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO smoothings DO BEGIN
               FOR j := 1 TO 3 DO BEGIN
                  vtot := q^.bigv + r^.bigv;
                  f1 := q^.bigv/vtot;
                  f2 := 1.0 - f1;
                  p^.bigv := (f1*r^.dist + f2*p^.dist - f1*f2*q^.dist)/df;
                  p^.bigv := p^.bigv - vtot*f1*f2;
                  IF p^.bigv < p^.back^.deltav
                  THEN p^.bigv := p^.back^.deltav;
                  temp := p;
                  p := q;
                  q := r;
                  r := temp;
                  END;
               END;
         END; (* correctv *)

         PROCEDURE littlev (p : nodeptr);
            VAR i : INTEGER;
         BEGIN (* littlev *)
            FOR i := 1 TO 3 DO BEGIN
               WITH p^ DO BEGIN
                  v := bigv - back^.deltav;
                  back^.v := v;
                  END;
               p := p^.next;
               END;
         END; (* littlev *)

         PROCEDURE nuview (p : nodeptr);
            VAR i, j : INTEGER;
               q, r, a, b, temp : nodeptr;
               v1, v2, vtot, f1, f2 : REAL;
         BEGIN (* nuview *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO 3 DO BEGIN
               a := q^.back;
               b := r^.back;
               v1 := q^.bigv;
               v2 := r^.bigv;
               vtot := v1 + v2;
               f1 := v2/vtot;
               f2 := v1/vtot;
               FOR j := 1 TO totalleles DO
                  p^.view[j] := f1*a^.view[j] + f2*b^.view[j];
               p^.deltav := v1*v2/vtot;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* nuview *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN BEGIN
            makedists (p);
            makebigv (p);
            IF negatives THEN correctv (p);
            littlev (p);
            nuview (p);
            END;
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
      BEGIN (* smooth *)
         IF NOT p^.tip
         THEN BEGIN
            update (p);
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE insert (p, q: nodeptr);
         VAR i : INTEGER;
      BEGIN (* insert *)
         hookup (p^.next^.next, q^.back);
         hookup (p^.next, q);
         FOR i :=1 TO smoothings DO BEGIN
            smooth (p);
            smooth (p^.back);
            END;
      END; (* insert *)

      PROCEDURE copy ( VAR a, b : tree);
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE setuptip (m : INTEGER; VAR t : tree);
      BEGIN (* setuptip *)
         WITH t.nodep[m]^ DO BEGIN
            view := x[m];
            nayme := names[m];
            deltav := 0.0;
            v := 0.0;
            END;
      END; (* setuptip *)

      PROCEDURE buildnewtip (m : INTEGER; VAR t : tree);
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         setuptip (m, t);
         p := t.nodep[m + numsp - 2];
         hookup (t.nodep[m], p);
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR t : tree);
      BEGIN (* buildsimpletree *)
         setuptip (1, t);
         setuptip (2, t);
         hookup (t.nodep[1], t.nodep[2]);
         buildnewtip (3, t);
         insert(t.nodep[3]^.back, t.nodep[1]);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr);
      BEGIN (* addtraverse *)
         insert(p, q);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN copy (curtree, bestree);
         copy (priortree, curtree);
         IF NOT q^.tip
         THEN BEGIN
            addtraverse (p, q^.next^.back);
            addtraverse (p, q^.next^.next^.back);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p, q : nodeptr);

         PROCEDURE remove ( VAR p, q : nodeptr);
         BEGIN (* remove *)
            q := p^.next^.back;
            hookup (q, p^.next^.next^.back);
            p^.next^.back := NIL;
            p^.next^.next^.back := NIL;
            update (q);
            update (q^.back);
         END; (* remove *)

         PROCEDURE tryrearr (p, q : nodeptr);
            VAR r : nodeptr;
         BEGIN (* tryrearr *)
            r := q;
            remove (p, r);
            insert (p, q);
            numtrees := numtrees + 1;
            IF evaluate(curtree) > bestree.likelihood
            THEN BEGIN
               succeeded := TRUE;
               copy (curtree, bestree);
               END
            ELSE copy (bestree, curtree);
         END; (* tryrearr *)

      BEGIN (* rearrange *)
         IF NOT succeeded
         THEN BEGIN
            IF (NOT p^.tip) AND (NOT q^.tip)
            THEN BEGIN
               tryrearr (p, q^.next);
               tryrearr (p, q^.next^.next);
               END;
            IF NOT q^.tip
            THEN BEGIN
               rearrange (q, q^.next^.back);
               rearrange (q, q^.next^.next^.back);
               END;
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + over;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.next^.back);
               p^.xcoord := p^.next^.back^.xcoord;
               coordinates (p^.next^.next^.back);
               p^.ycoord := p^.next^.next^.back^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            WRITE(OUTPUT, ' ');
            REPEAT
               IF p^.next^.back^.ycoord >= i
               THEN p := p^.next^.back
               ELSE BEGIN
                  q := p^.next^.next^.back;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF p^.number > 9
                     THEN WRITE(OUTPUT, p^.number:2)
                     ELSE WRITE(OUTPUT, p^.number:1, '-');
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO namelength DO WRITE(OUTPUT, p^.nayme[j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         coordinates (curtree.start^.back);
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO tipy-down DO drawline (i);
         FOR i := 1 TO down DO WRITELN(OUTPUT, ' !');
         WRITE(OUTPUT, ' ');
         FOR i := 1 TO namelength DO
            WRITE(OUTPUT, curtree.start^.nayme[i]);
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE summarize;
         VAR chilow, chihigh : REAL;

         PROCEDURE describe (p : nodeptr);
            VAR i : INTEGER;
               q : nodeptr;
               bigv, delta : REAL;
         BEGIN (* describe *)
            q := p^.back;
            WRITE(OUTPUT, ' ');
            IF q^.tip
            THEN FOR i := 1 TO namelength DO WRITE(OUTPUT, q^.nayme[i])
            ELSE WRITE(OUTPUT, q^.number:4, '      ');
            WRITE(OUTPUT, '    ');
            IF p^.tip
            THEN FOR i := 1 TO namelength DO WRITE(OUTPUT, p^.nayme[i])
            ELSE WRITE(OUTPUT, p^.number:4, '      ');
            WRITE(OUTPUT, q^.v:15:5);
            delta := p^.deltav + p^.back^.deltav;
            bigv := p^.v + delta;
            WRITELN(OUTPUT, '   (', chilow*bigv-delta:12:5, ',',
               chihigh*bigv-delta:12:5,')');
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT, ' remember: ');
         IF didreroot
         THEN WRITE(OUTPUT, '(although rooted by outgroup) ');
         WRITELN(OUTPUT, 'this is an unrooted tree!');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Ln Likelihood = ', curtree.likelihood:11:5);
         IF NOT usertree
         THEN BEGIN
            WRITELN(OUTPUT, ' ');
            WRITELN(OUTPUT, ' Examined ',numtrees:4, ' trees');
            END;
         IF df = 1
         THEN BEGIN
            chilow := 0.000982;
            chihigh := 5.02389;
            END
         ELSE IF df = 2
            THEN BEGIN
               chilow := 0.05064;
               chihigh := 7.3777;
               END
            ELSE BEGIN
               chilow := 1.0 - 2.0/(9*df);
               chihigh := chilow;
               chilow := chilow - 1.95996 * SQRT(2.0/(9*df));
               chihigh := chihigh + 1.95996 * SQRT(2.0/(9*df));
               chilow := chilow * chilow * chilow;
               chihigh := chihigh * chihigh * chihigh;
               END;
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT, '  Between        And            Length');
         WRITELN(OUTPUT, '      Approx. Confidence Limits');
         WRITE(OUTPUT, '  -------        ---            ------');
         WRITELN(OUTPUT, '      ------- ---------- ------');
         describe (curtree.start^.back);
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' ');
      END; (* summarize *)

      PROCEDURE treeread;
         VAR ch : CHAR;
            p : nodeptr;
            i, nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            VAR q : nodeptr;
               n : INTEGER;
         BEGIN (* addelement *)
            REPEAT
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := curtree.nodep[nextnode];
               hookup(p, q);
               addelement (q^.next);
               findch (',');
               addelement (q^.next^.next);
               findch (')');
               READ(infile, ch);
               END
            ELSE BEGIN
               n := 0;
               REPEAT
                  n := 10*n + ORD(ch) - ORD('0');
                  READ(infile, ch);
               UNTIL NOT((ORD(ch)>=ORD('0')) AND (ORD(ch)<=ORD('9')));
               hookup (curtree.nodep[n], p);
               IF curtree.start^.number > n
               THEN curtree.start := curtree.nodep[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               addelement (p);
               p := p^.next;
               findch (',');
               END;
            addelement (p);
            FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
            END;
         READLN(infile);
      END; (* treeread *)

      PROCEDURE treevaluate;
         VAR p : nodeptr;
            dummy : REAL;

         PROCEDURE nodeinit (p : nodeptr);
            VAR q, r : nodeptr;
               i : INTEGER;
         BEGIN (* nodeinit *)
            IF NOT p^.tip
            THEN BEGIN
               q := p^.next^.back;
               r := p^.next^.next^.back;
               nodeinit (q);
               nodeinit (r);
               FOR i := 1 TO totalleles DO
                  p^.view[i] := 0.5 * q^.view[i] + 0.5 * r^.view[i];
               END;
         END; (* nodeinit *)

      BEGIN (* treevaluate *)
         nodeinit (curtree.start);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      IF usertree
      THEN BEGIN
         setuptree (curtree);
         FOR which := 1 TO numsp DO setuptip (which, curtree);
         READLN(infile, numtrees);
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            treevaluate;
            printree;
            summarize;
            END;
         END
      ELSE BEGIN
         setuptree (curtree);
         setuptree (priortree);
         setuptree (bestree);
         buildsimpletree (curtree);
         numtrees := 1;
         nextsp := 4;
         WHILE nextsp <= numsp DO BEGIN
            buildnewtip (nextsp, curtree);
            copy (curtree, priortree);
            bestree.likelihood := -99999.0;
            addtraverse (curtree.nodep[nextsp]^.back,
                                  curtree.start^.back);
            copy (bestree, curtree);
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start, curtree.start^.back);
               END;
            IF (NOT brief) OR (nextsp = numsp)
            THEN BEGIN
               didreroot := (nextsp = numsp);
               IF didreroot
               THEN curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               END;
            nextsp := nextsp + 1;
            END;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   RESET(infile);
   getinput;
   maketree;
END.
