PROGRAM Fitch (infile, OUTPUT);
   (* version 2.5 *)

CONST maxsp = 10;   (* maximum number of species *)
   maxsp2 = 18;     (* 2 x maxsp - 2 *)
   smoothings = 4;  (* number of passes through smoothing algorithm *)
   namelength = 10; (* number of characters max. in species name *)

TYPE vector = ARRAY [1..maxsp2] OF REAL;
   naym = ARRAY[1..namelength] OF CHAR;

VAR infile : TEXT;
   numsp, numsp1, numsp2, numchar, power, nums, outgrno : INTEGER;
   x : ARRAY[1..maxsp] OF vector;
   names : ARRAY[1..maxsp] OF naym;
   brief, usertree, lower, upper, negallowed,
     powers, outgropt, didreroot : BOOLEAN;

   PROCEDURE getinput;

      PROCEDURE getnums;
      BEGIN (* getnums *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT,' Fitch-Margoliash method,');
         WRITELN(OUTPUT, ' version 2.5');
         WRITELN(OUTPUT, ' ');
         READ(infile, numsp);
         WRITELN(OUTPUT, ' ', numsp:4, ' Populations');
         numsp1 := numsp + 1;
         numsp2 := 2*numsp - 2;
      END; (* getnums *)

      PROCEDURE getoptions;
         VAR ch : CHAR;
            i, extranum : INTEGER;
      BEGIN (* getoptions *)
         usertree := FALSE;
         lower := FALSE;
         upper := FALSE;
         brief := FALSE;
         negallowed := FALSE;
         outgropt := FALSE;
         powers := FALSE;
         power := 2;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            IF ch <> ' '
            THEN BEGIN
               IF ORD(ch) > ORD('Z')
               THEN ch := CHR(ORD(ch)+ORD('Z')-ORD('z'));
               CASE ch OF
                  'B' : brief := TRUE;
                  'L' : lower := TRUE;
                  'N' : negallowed := TRUE;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'P' : BEGIN
                        powers := TRUE;
                        extranum := extranum + 1;
                        END;
                  'R' : upper := TRUE;
                  'U' : usertree := TRUE
                  END;
               END;
            END;
         FOR i := 1 TO extranum DO BEGIN
            READLN(infile);
            READ(infile, ch);
            IF ORD(ch) > ORD('Z')
            THEN ch := CHR(ORD(ch)+ORD('Z')-ORD('z'));
            CASE ch OF
               'O' : READ(infile, outgrno);
               'P' : READ(infile, power)
               END;
            END;
         WRITELN(OUTPUT, '                                     2');
         WRITELN(OUTPUT, '                   <  <   (Obs - Exp)');
         WRITELN(OUTPUT, ' Sum of squares =  <  <   ------------');
         WRITELN(OUTPUT, '                   i  j         ',
                        power:2);
         WRITELN(OUTPUT, '                              Obs');
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Negative branch lengths ');
         IF NOT negallowed
         THEN WRITE(OUTPUT, 'not ');
         WRITELN(OUTPUT, 'allowed');
      END; (* getoptions *)

      PROCEDURE getdata;
         VAR i, j, k : INTEGER;
            skipit, skipother : BOOLEAN;
      BEGIN (* getdata *)
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' Name                       Distances');
         WRITELN(OUTPUT, ' ----                       ---------');
         WRITELN(OUTPUT);
         FOR i := 1 TO numsp DO BEGIN
            x[i, i] := 0.0;
            READLN(infile);
            FOR j := 1 TO namelength DO READ(infile, names[i, j]);
            FOR j := 1 TO numsp DO BEGIN
               skipit := (lower AND (j >= i)) OR (upper AND (j <= i));
               skipother := (lower AND (i >= j)) OR
                              (upper AND (i <= j));
               IF NOT skipit
               THEN BEGIN
                  IF EOLN(infile)
                  THEN READLN(infile);
                  READ(infile, x[i, j]);
                  END;
               IF (NOT skipit) AND skipother THEN x[j, i] := x[i, j];
               END;
            END;
         FOR i := 1 TO numsp DO BEGIN
            WRITE(OUTPUT, ' ');
            FOR j := 1 TO namelength DO WRITE(OUTPUT, names[i,j]);
            WRITE(OUTPUT, ' ');
            FOR j := 1 TO numsp DO BEGIN
               WRITE(OUTPUT, x[i, j]:10:5);
               IF ((j MOD 6) = 0) AND (j < numsp)
               THEN BEGIN
                  WRITELN(OUTPUT);
                  FOR k := 1 TO namelength+4 DO
                     WRITE(OUTPUT, ' ');
                  END;
               END;
            WRITELN(OUTPUT);
            END;
      END; (* getdata *)

   BEGIN (* getinput *)
      getnums;
      getoptions;
      getdata;
   END; (* getinput *)

   PROCEDURE maketree;

      TYPE nodeptr = ^node;
         node = RECORD
            next, back : nodeptr;
            tip        : BOOLEAN;
            number     : INTEGER;
            nayme      : naym;
            d, w       : vector;
            v          : REAL;
            dist       : REAL;
            xcoord,
             ycoord    : INTEGER
            END;
         tree = RECORD
            nodep      : ARRAY[1..maxsp2] OF nodeptr;
            likelihood : REAL;
            start      : nodeptr
            END;

      VAR curtree, priortree, bestree : tree;
         nextsp, numtrees, which : INTEGER;
         succeeded : BOOLEAN;

      PROCEDURE setuptree (VAR a : tree);
        VAR i, j : INTEGER;
           p, q : nodeptr;
      BEGIN (* setuptree *)
         WITH a DO BEGIN
            FOR i := 1 TO numsp DO BEGIN
               NEW(nodep[i]);
               nodep[i]^.tip := TRUE;
               nodep[i]^.number := i;
               END;
            FOR i := numsp1 TO numsp2 DO BEGIN
               q := NIL;
               FOR j := 1 TO 3 DO BEGIN
                  NEW(p);
                  p^.tip := FALSE;
                  p^.number := i;
                  p^.next := q;
                  q := p;
                  END;
               p^.next^.next^.next := p;
               nodep[i] := p;
               END;
            likelihood := -1.0;
            start := nodep[1];
            END;
      END; (* setuptree *)

      PROCEDURE hookup (p, q : nodeptr);
      BEGIN (* hookup *)
         p^.back := q;
         q^.back := p;
      END; (* hookup *)

      FUNCTION evaluate (VAR t : tree ) : REAL;
         VAR sum : REAL;

         PROCEDURE firsttraverse (p : nodeptr);
            VAR nx : INTEGER;

            PROCEDURE secondtraverse (q : nodeptr; y : REAL);
               VAR z : REAL;
            BEGIN (* secondtraverse *)
               z := y + q^.v;
               IF q^.tip
               THEN sum := sum + q^.w[nx] * SQR(q^.d[nx] - z)
               ELSE BEGIN
                  secondtraverse (q^.next^.back, z);
                  secondtraverse (q^.next^.next^.back, z);
                  END;
            END; (* secondtraverse *)

         BEGIN (* firsttraverse *)
            IF p^.tip
            THEN BEGIN
               nx := p^.number;
               secondtraverse (p^.back, 0.0);
               END
            ELSE BEGIN
               firsttraverse (p^.next^.back);
               firsttraverse (p^.next^.next^.back);
               END;
         END; (* firsttraverse *)

      BEGIN (* evaluate *)
         sum := 0.0;
         firsttraverse (t.start^.back^.back);
         firsttraverse (t.start^.back);
         t.likelihood := -sum;
         evaluate := -sum;
      END; (* evaluate *)

      PROCEDURE nudists (x, y : nodeptr);
         VAR nx, ny : INTEGER;
            dil, djl, wil, wjl, vi, vj : REAL;
            qprime, rprime : nodeptr;
      BEGIN (* nudists *)
         qprime := x^.next;
         rprime := qprime^.next;
         qprime := qprime^.back;
         rprime := rprime^.back;
         nx := x^.number;
         ny := y^.number;
         dil := qprime^.d[ny];
         djl := rprime^.d[ny];
         wil := qprime^.w[ny];
         wjl := rprime^.w[ny];
         vi := qprime^.v;
         vj := rprime^.v;
         x^.w[ny] := wil + wjl;
         y^.w[nx] := x^.w[ny];
         x^.d[ny] := ((dil - vi) * wil + (djl - vj) * wjl)
                          /(wil + wjl);
         y^.d[nx] := x^.d[ny];
      END; (* nudists *)

      PROCEDURE update (p : nodeptr);
         VAR q : nodeptr;
            negatives : BOOLEAN;

         PROCEDURE makedists (p : nodeptr);
            VAR i, n : INTEGER;
               q : nodeptr;
         BEGIN (* makedists *)
            FOR i := 1 TO 3 DO BEGIN
               n := p^.back^.number;
               q := p^.next;
               p^.dist := q^.back^.d[n];
               p := q;
               END;
         END; (* makedists *)

         PROCEDURE makebigv (p : nodeptr);
            VAR i : INTEGER;
               temp, q, r : nodeptr;
         BEGIN (* makebigv *)
            q := p^.next;
            r := q^.next;
            negatives := FALSE;
            FOR i := 1 TO 3 DO BEGIN
               p^.v := (p^.dist + r^.dist - q^.dist)/2.0;
               IF p^.v < 0.0 THEN negatives := TRUE;
               p^.back^.v := p^.v;
               temp := p;
               p := q;
               q := r;
               r := temp;
               END;
         END; (* makebigv *)

         PROCEDURE correctv (p : nodeptr);
            VAR q, r, temp : nodeptr;
               i, j, n : INTEGER;
               wq, wr : REAL;
         BEGIN (* correctv *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO smoothings DO
               FOR j := 1 TO 3 DO BEGIN
                  n := p^.back^.number;
                  wr := r^.back^.w[n];
                  wq := q^.back^.w[n];
                  p^.v := ((p^.dist - q^.v) * wq
                           + (r^.dist - r^.v) * wr)
                            / (wr + wq);
                  IF p^.v < 0 THEN p^.v := 0.0;
                  p^.back^.v := p^.v;
                  temp := p;
                  p := q;
                  q := r;
                  r := temp;
                  END;
         END; (* correctv *)

         PROCEDURE nuview (p : nodeptr);
            VAR i : INTEGER;
               q, r, pprime, qprime, rprime, temp : nodeptr;

            PROCEDURE alter (x, y : nodeptr);
            BEGIN (* alter *)
               nudists (x, y);
               IF NOT y^.tip
               THEN BEGIN
                  alter(x, y^.next^.back);
                  alter(x, y^.next^.next^.back);
                  END;
            END; (* alter *)

         BEGIN (* nuview *)
            q := p^.next;
            r := q^.next;
            FOR i := 1 TO 3 DO BEGIN
               temp := p;
               pprime := p^.back;
               qprime := q^.back;
               rprime := r^.back;
               alter (p, pprime);
               p := q;
               q := r;
               r := temp;
               END;
         END; (* nuview *)

      BEGIN (* update *)
         IF NOT p^.tip
         THEN BEGIN
            makedists (p);
            makebigv (p);
            IF (negatives) AND (NOT negallowed) THEN correctv (p);
            nuview (p);
            END;
      END; (* update *)

      PROCEDURE smooth (p : nodeptr);
      BEGIN (* smooth *)
         IF NOT p^.tip
         THEN BEGIN
            update (p);
            smooth (p^.next^.back);
            smooth (p^.next^.next^.back);
            END;
      END; (* smooth *)

      PROCEDURE insert (p, q: nodeptr; continue : BOOLEAN);
         VAR i : INTEGER;
            x : REAL;

         PROCEDURE fillin (pa, qa : nodeptr);

            PROCEDURE filltraverse (pb, qb : nodeptr);
            BEGIN (* filltraverse *)
               IF NOT qb^.tip
               THEN IF continue
                  THEN BEGIN
                     filltraverse (pb, qb^.next^.back);
                     filltraverse (pb, qb^.next^.next^.back);
                     nudists (qb, pb);
                     END
                  ELSE BEGIN
                     IF NOT qb^.next^.back^.tip
                     THEN nudists (qb^.next^.back, pb);
                     IF NOT qb^.next^.next^.back^.tip
                     THEN nudists (qb^.next^.next^.back, pb);
                     END;
            END; (* filltraverse *)

         BEGIN (* fillin *)
            IF NOT pa^.tip
            THEN BEGIN
              fillin (pa^.next^.back, qa);
              fillin (pa^.next^.next^.back, qa);
              END;
            filltraverse (pa, qa);
         END; (* fillin *)

      BEGIN (* insert *)
         hookup (p^.next^.next, q^.back);
         hookup (p^.next, q);
         x := q^.v/2.0;
         p^.v := 0.0;
         p^.back^.v := 0.0;
         p^.next^.v := x;
         p^.next^.back^.v := x;
         p^.next^.next^.back^.v := x;
         p^.next^.next^.v := x;
         fillin (p^.back, p);
         FOR i :=1 TO smoothings DO BEGIN
            smooth (p);
            smooth (p^.back);
            END;
      END; (* insert *)

      PROCEDURE copy ( VAR a, b : tree);
         VAR i : INTEGER;

         PROCEDURE copynode( VAR c, d : nodeptr);
            VAR necks : nodeptr;
         BEGIN (* copynode *)
            necks := d^.next;
            d^ := c^;
            d^.next := necks;
         END; (* copynode *)

      BEGIN (* copy *)
         FOR i := 1 TO numsp DO b.nodep[i]^ := a.nodep[i]^;
         FOR i := numsp1 TO numsp2 DO BEGIN
            copynode (a.nodep[i], b.nodep[i]);
            copynode (a.nodep[i]^.next, b.nodep[i]^.next);
            copynode (a.nodep[i]^.next^.next, b.nodep[i]^.next^.next);
            END;
         b.likelihood := a.likelihood;
         b.start := a.start;
      END; (* copy *)

      PROCEDURE setuptip (m : INTEGER; VAR t : tree);
         VAR i, j : INTEGER;
      BEGIN (* setuptip *)
         WITH t.nodep[m]^ DO BEGIN
            d := x[m];
            FOR i := 1 TO numsp DO
               IF i <> m
               THEN BEGIN
                  w[i] := 1.0;
                  FOR j := 1 TO power DO
                     w[i] := w[i] / d[i];
                  END
               ELSE BEGIN
                  w[m] := 0.0;
                  d[m] := 0.0;
                  END;
            number := m;
            nayme := names[m];
            v := 0.0;
            END;
      END; (* setuptip *)

      PROCEDURE buildnewtip (m : INTEGER; VAR t : tree);
         VAR p : nodeptr;
      BEGIN (* buildnewtip *)
         setuptip (m, t);
         p := t.nodep[m + numsp - 2];
         hookup (t.nodep[m], p);
      END; (* buildnewtip *)

      PROCEDURE buildsimpletree (VAR t : tree);
      BEGIN (* buildsimpletree *)
         setuptip (1, t);
         setuptip (2, t);
         hookup (t.nodep[1], t.nodep[2]);
         buildnewtip (3, t);
         insert(t.nodep[3]^.back, t.nodep[1], FALSE);
      END; (* buildsimpletree *)

      PROCEDURE addtraverse (p, q : nodeptr);
      BEGIN (* addtraverse *)
         insert(p, q, TRUE);
         numtrees := numtrees + 1;
         IF evaluate(curtree) > bestree.likelihood
         THEN copy (curtree, bestree);
         copy (priortree, curtree);
         IF NOT q^.tip
         THEN BEGIN
            addtraverse (p, q^.next^.back);
            addtraverse (p, q^.next^.next^.back);
            END;
      END; (* addtraverse *)

      PROCEDURE rearrange (p, q : nodeptr);

         PROCEDURE remove ( VAR p, q : nodeptr);
         BEGIN (* remove *)
            q := p^.next^.back;
            hookup (q, p^.next^.next^.back);
            p^.next^.back := NIL;
            p^.next^.next^.back := NIL;
            update (q);
            update (q^.back);
         END; (* remove *)

         PROCEDURE tryrearr (p, q : nodeptr);
            VAR r : nodeptr;
         BEGIN (* tryrearr *)
            r := q;
            remove (p, r);
            insert (p, q, FALSE);
            numtrees := numtrees + 1;
            IF evaluate(curtree) > bestree.likelihood
            THEN BEGIN
               succeeded := TRUE;
               copy (curtree, bestree);
               END
            ELSE copy (bestree, curtree);
         END; (* tryrearr *)

      BEGIN (* rearrange *)
         IF NOT succeeded
         THEN BEGIN
            IF (NOT p^.tip) AND (NOT q^.tip)
            THEN BEGIN
               tryrearr (p, q^.next);
               tryrearr (p, q^.next^.next);
               END;
            IF NOT q^.tip
            THEN BEGIN
               rearrange (q, q^.next^.back);
               rearrange (q, q^.next^.next^.back);
               END;
            END;
      END; (* rearrange *)

      PROCEDURE printree;
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i : INTEGER;

         PROCEDURE coordinates (p : nodeptr);
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + over;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.next^.back);
               p^.xcoord := p^.next^.back^.xcoord;
               coordinates (p^.next^.next^.back);
               p^.ycoord := p^.next^.next^.back^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            VAR p, q : nodeptr;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p := curtree.start^.back;
            WRITE(OUTPUT, ' ');
            REPEAT
               IF p^.next^.back^.ycoord >= i
               THEN p := p^.next^.back
               ELSE BEGIN
                  q := p^.next^.next^.back;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF p^.number > 9
                     THEN WRITE(OUTPUT, p^.number:2)
                     ELSE WRITE(OUTPUT, p^.number:1, '-');
                     FOR j := 1 TO n-2 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO namelength DO
               WRITE(OUTPUT, p^.nayme[j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         coordinates (curtree.start^.back);
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO tipy-down DO drawline (i);
         FOR i := 1 TO down DO WRITELN(OUTPUT, ' !');
         WRITE(OUTPUT, ' ');
         FOR i := 1 TO namelength DO WRITE(OUTPUT,
            curtree.start^.nayme[i]);
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE summarize;

         PROCEDURE describe (p : nodeptr);
            VAR i : INTEGER;
               q : nodeptr;
               bigv, delta : REAL;
         BEGIN (* describe *)
            q := p^.back;
            WRITE(OUTPUT, ' ');
            IF q^.tip
            THEN FOR i := 1 TO namelength DO
               WRITE(OUTPUT, q^.nayme[i])
            ELSE WRITE(OUTPUT, q^.number:4, '      ');
            WRITE(OUTPUT, '    ');
            IF p^.tip
            THEN FOR i := 1 TO namelength DO
               WRITE(OUTPUT, p^.nayme[i])
            ELSE WRITE(OUTPUT, p^.number:4, '      ');
            WRITE(OUTPUT, q^.v:15:5);
            WRITELN(OUTPUT);
            IF NOT p^.tip
            THEN BEGIN
               describe (p^.next^.back);
               describe (p^.next^.next^.back);
               END;
         END; (* describe *)

      BEGIN (* summarize *)
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT, ' remember:');
         IF didreroot
         THEN WRITE(OUTPUT, ' (although rooted by outgroop)');
         WRITELN(OUTPUT, ' this is an unrooted tree!');
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' Sum of squares = ',
                     -curtree.likelihood:11:5);
         WRITELN(OUTPUT, ' ');
         IF power = 2
         THEN BEGIN
            WRITE(OUTPUT, ' Average percent standard deviation = ');
            WRITELN(OUTPUT, 100*SQRT(-curtree.likelihood/
                              (nums*(nums-1)-2)):11:5);
            WRITELN(OUTPUT, ' ');
            END;
         IF NOT usertree
         THEN BEGIN
            WRITELN(OUTPUT, ' ');
            WRITELN(OUTPUT, ' examined ',numtrees:4, ' trees');
            END;
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, '  Between        And            Length');
         WRITELN(OUTPUT, '  -------        ---            ------');
         describe (curtree.start^.back);
         WRITELN(OUTPUT, ' ');
         WRITELN(OUTPUT, ' ');
      END; (* summarize *)

      PROCEDURE treeread;
         VAR ch : CHAR;
            p : nodeptr;
            i, nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (p : nodeptr);
            VAR q : nodeptr;
               n : INTEGER;
         BEGIN (* addelement *)
            REPEAT
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := curtree.nodep[nextnode];
               hookup(p, q);
               addelement (q^.next);
               findch (',');
               addelement (q^.next^.next);
               findch (')');
               READ(infile, ch);
               END
            ELSE BEGIN
               n := 0;
               REPEAT
                  n := 10*n + ORD(ch) - ORD('0');
                  READ(infile, ch);
               UNTIL NOT((ORD(ch)>=ORD('0')) AND (ORD(ch)<=ORD('9')));
               nums := nums + 1;
               hookup (curtree.nodep[n], p);
               IF curtree.start^.number > n
               THEN curtree.start := curtree.nodep[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         nums := 0;
         curtree.start := curtree.nodep[numsp];
         REPEAT
            READ(infile, ch);
         UNTIL ch <> ' ';
         IF ch = '('
         THEN BEGIN
            nextnode := numsp + 1;
            p := curtree.nodep[nextnode];
            FOR i := 1 TO 2 DO BEGIN
               addelement (p);
               p := p^.next;
               findch (',');
               END;
            addelement (p);
            END;
         READLN(infile);
      END; (* treeread *)

      PROCEDURE treevaluate;
         VAR p : nodeptr;
            i : INTEGER;
            dummy : REAL;

         PROCEDURE nodeinit (p : nodeptr);
            VAR i, j : INTEGER;
         BEGIN (* nodeinit *)
            IF p^.tip
            THEN FOR i := numsp1 TO numsp2 DO BEGIN
               p^.w[i] := 1.0;
               p^.d[i] := 0.0;
               END
            ELSE BEGIN
               FOR i := 1 TO 3 DO BEGIN
                  FOR j := 1 TO numsp2 DO BEGIN
                     p^.w[j] := 1.0;
                     p^.d[j] := 0.0;
                     END;
                  p := p^.next;
                  END;
               END;
          END; (* nodeinit *)

      BEGIN (* treevaluate *)
         FOR i := 1 TO numsp2 DO
            nodeinit(curtree.nodep[i]);
         FOR i := 1 TO 4*smoothings DO smooth (curtree.start^.back);
         dummy := evaluate (curtree);
      END; (* treevaluate *)

   BEGIN (* maketree *)
      IF usertree
      THEN BEGIN
         setuptree (curtree);
         FOR which := 1 TO numsp DO setuptip (which, curtree);
         READLN(infile, numtrees);
         FOR which := 1 TO numtrees DO BEGIN
            treeread;
            treevaluate;
            printree;
            summarize;
            END;
         END
      ELSE BEGIN
         setuptree (curtree);
         setuptree (priortree);
         setuptree (bestree);
         buildsimpletree (curtree);
         numtrees := 1;
         nextsp := 4;
         WHILE nextsp <= numsp DO BEGIN
            nums := nextsp;
            buildnewtip (nextsp, curtree);
            copy (curtree, priortree);
            bestree.likelihood := -99999.0;
            addtraverse (curtree.nodep[nextsp]^.back,
                                   curtree.start^.back);
            copy (bestree, curtree);
            succeeded := TRUE;
            WHILE succeeded DO BEGIN
               succeeded := FALSE;
               rearrange (curtree.start, curtree.start^.back);
               END;
            IF (NOT brief) OR (nextsp = numsp)
            THEN BEGIN
               didreroot := outgropt AND (nextsp = numsp);
               IF didreroot
               THEN curtree.start := curtree.nodep[outgrno];
               printree;
               summarize;
               END;
            nextsp := nextsp + 1;
            END;
         END;
   END; (* maketree *)

BEGIN (* main program *)
   RESET(infile);
   getinput;
   maketree;
END.
